/*********************************************************************/
/* read.c                                                            */
/* function neblo_handle_REQ_msgs()                                 */
/* see COPYRIGHT in parent directory                                 */
/*********************************************************************/

#include <neblo_.h>
#include <neblo_macros.h>
#include <neblo_secret.h>
#include <queues.h>


/*
 * this peer is asked to take a read msg, and deliver to application
 * if this peer is the final destination, or forward on otherwise
 */
int
neblo_handle_REQ_msgs
( reqst_t *reqst, unsigned int retry )
{

  unsigned int          rport;
  neblo_port            port;
  unsigned int          maxlen;
  neblo_rhandle         handle;
  neblo_addr            addr;
  neblo_addr            real_addr;
  peer_t               *from;
  int                   rc;
  peer_t               *hop;
  unsigned int          actual_len;
  pend_read_t          *pr;
  neblo_rhandle         new_h;
  unsigned int          tmpunavail = 0;


  sscanf(reqst->buf,"%*s %04x %08x %08x %016llx %016llx",
         &rport,&maxlen,&handle,&addr,&real_addr);
  port = (neblo_port)rport;

#ifdef DEBUG_RW
  if (retry != FIRST_ATTMPT)
    S(fprintf(stderr,
              "[%lu]\tRETRY read: port %u, addr %016llx, real_addr %016llx, maxlen %u\n",
              reqst->th_id,port,addr,real_addr,maxlen));
#endif

  from = is_friend(reqst->csock);

  /*
   * SECURITY: requests of this kind may only be accepted
   * if coming from a friend.
   * however, if this is a retry, the request might be from the
   * local peer; in this case it must be accepted
   */
  if (!from  && retry == FIRST_ATTMPT) {

#ifdef DEBUG_RW
    S(fprintf(stderr,"[%lu]\tREFUSED read: not a friend\n",reqst->th_id));
#endif

    return -1;
  }

  rc = 0;
  new_h = (neblo_rhandle)0;

  for (;;) {

    hop = next_hop(addr);

    if (hop == NULL) {
      /*
       * "addr" is mapped locally.
       * deliver msg to application, stripping header away
       */
      actual_len = maxlen;
      rc = invoke_receiver_handler_REQ(
             port,real_addr,reqst->buf+HEADERSIZE_R,&actual_len);

#ifdef DEBUG_RW
      S(fprintf(stderr,
                "[%lu]\t``read'' to port %u addr %016llx (real addr %016llx) returned %d, len %u, msg %s\n",
                reqst->th_id,port,addr,real_addr,rc,actual_len,
                reqst->buf+HEADERSIZE_R));
#endif

      if (rc < 0)
        /* application handler encountered troubles (missing data?) */
        break;

      /*
       * this is a read msg, so the application handler may have changed
       * the content of mem region at address reqst->buf+HEADERSIZE_R
       * and has set the value for "actual_len" to the actual data size
       */

      actual_len = min(maxlen,actual_len);

      /*
       *  Now send the answer, backtracking the route.
       *  To avoid a memcpy, overwrite old header with new one; this is
       *  feasible as the two headers have same size.
       */

      /*
       *  If the read msg originated locally (retried operation)
       *  ("from" = NULL, and therefore retry != FIRST_ATTMPT)
       *  then the answer must be passed up to the application as there is no
       *  actual route to backtrack.
       */

      if (from != NULL) {
        rc = forward_msg(DATA,from,port,addr,real_addr,reqst->buf,actual_len,
                         handle,0,reqst->th_id);
        if (rc >= 0)
          /* success. quit service */
          break;

#ifdef DEBUG_RW
        S(fprintf(stderr,
                  "[%lu]\tforward_msg() to addr %016llx (real_addr %016llx) returned %d\n",
                  reqst->th_id,addr,real_addr,rc));
#endif

        break;
      }

      /*  Locally-originated retried read...  */
      pr = search_read_reqst(handle);
      if (pr == NULL) {
        /*  read handle lost meanwhile.  should not happen, but anyway... */
        rc = -1;
        break;
      }
      pthread_mutex_lock(&pr->lock);
      if (pr->done > 0) {
        /* read msg already successfully served, so discard and quit */
        pthread_mutex_unlock(&pr->lock);
        rc = 0;
        break;
      }
      /* trunc to max allowed size for this read msg */
      actual_len = min(actual_len,pr->maxlen);
      if (actual_len) {
        /* pass answer up to application, then quit */
        memcpy(pr->data,reqst->buf+HEADERSIZE_R,actual_len);
        pr->done = 1;
        pthread_cond_signal(&pr->wait_cond);
        /* neblo_RES_wait()/neblo_RES_test() will cancel the request from queue */
      } else {
        /* zero-sized data means "data not found" */
        pr->done = -1;
      }
      pthread_mutex_unlock(&pr->lock);
      rc = 0;
      break;
    }

    if (hop == (peer_t *)(-1)) {
      /*
       * "addr" is not mapped here, and this peer doesn't know where to
       * route this msg copy at the moment.
       */

#ifdef DEBUG_RW
      if (retry != LAST_CHANCE)
        S(fprintf(stderr,
                  "[%lu]\tread delayed: port %u, addr %016llx, real_addr %016llx, maxlen %u\n",
                  reqst->th_id,port,addr,real_addr,maxlen));
      else
        S(fprintf(stderr,
                  "[%lu]\tread DISCARDED: port %u, addr %016llx, real_addr %016llx, maxlen %u\n",
                  reqst->th_id,port,addr,real_addr,maxlen));
#endif
      /*
       * insert the requested operation in a queue, to be tried again later,
       * but only if this is not a retry in itself (otherwise we could be
       * inserting the operation multiple times)
       */
      if (retry == FIRST_ATTMPT) {
        /*
         * we must do a copy. the original reqst buffer will be put back
         * in the service queue, the copy is to be put in the retry queue
         */
        reqst_t *r;

        _do_(_malloc_(r,sizeof(reqst_t)),!r,
             "neblo_handle_REQ_msgs(): r malloc");
        memcpy(r,reqst,sizeof(reqst_t));
        _do_(_calloc_(r->buf,KEYBUFLEN+SLACK),!r->buf,
             "neblo_handle_REQ_msgs(): r->buf calloc");
        memcpy(r->buf,reqst->buf,HEADERSIZE_R+maxlen);
        insert_retry_queue(READ,r,reqst->th_id);
      }
      else
        if (retry == LAST_CHANCE) {
          /* no more chances.  Msg is dropped :-( */
        }

      rc = -1;
      break;
    }

    /* "addr" is mapped remotely; route msg through peer "hop" */

    rc = invoke_forwarder_handler_REQ(port,real_addr,reqst->buf+HEADERSIZE_R,
                                      maxlen);
    if (rc < 0)
      /*  application decided not to forward this msg  */
      break;

    /*
     *  Take note of this pending read msg being forwarded, get a rhandle
     *  into "new_h", then forward the read msg to next hop.
     *  Retried ops have already been taken note when first attempted.
     */
    if (retry == FIRST_ATTMPT) {
      rc = insert_pend_read(port,real_addr,NULL,maxlen,from,handle,&new_h);
      if (rc < 0)
        /*  queue of pending read msgs full. service failed.  */
        break;
      rc = forward_msg(READ,hop,port,addr,real_addr,reqst->buf,maxlen,new_h,
                       1,reqst->th_id);
    } else
      /*  For retried ops, use the original handle to stamp the msg.  */
      rc = forward_msg(READ,hop,port,addr,real_addr,reqst->buf,maxlen,handle,
                       1,reqst->th_id);

    if (rc >= 0)
      /* success. quit service */
      break;

    /*
     * something wrong while talking to the selected peer, might have
     * been flagged as "tmpunavail"
     */

#ifdef DEBUG_RW
    S(fprintf(stderr,
              "[%lu]\tforward_msg() to addr %016llx (real_addr %016llx) returned %d\n",
              reqst->th_id,addr,real_addr,rc));
#endif

    tmpunavail = 1;
    /* choose another peer */

  }

  /*
   * the forward_msg() routine above might have flagged some peers as
   * "tmpunavail"...
   */
  if (tmpunavail)
    forall_peers(f,f->tmpunavail = 0);

  return rc;

}
