/*********************************************************************/
/* register.c                                                        */
/* function neblo_register()                                         */
/* see COPYRIGHT in parent directory                                 */
/*********************************************************************/

#include <neblo_.h>
#include <neblo_macros.h>
#include <neblo_secret.h>
#include <queues.h>


/****************************************************************************

connect algorithm, phase 1 (T is the foreign node, S is this peer; S is seen
as a ``seed'' from T).
it is based on the use of crypto hash functions on 160 bit (RIPEMD160):
T and S each invent a (64 bit) random number, resp. A and B;
T sends RIPEMD160(A) to S;
S sends RIPEMD160(B) to T;
T sends A to S (so S can prove fairness of T);
S sends B to T (so T can prove fairness of S);
SRC = (A xor B) (performed by both T and S);
C = RIPEMD160(SRC) (XOR-folded over 64 bit) (performed by both T and S);
S stores the join pending request from T.
C is the identifier for T, it will be used in phase 2 of the connect algorithm
(see "join" command after).
SRC is the "source" of C, and must be provided along with C when contacting
other peers in the 2nd phase of connect.
As C is the outcome of a RIPEMD160 (and SRC is provided along with C as a
proof of this), it is very difficult to forge a predetermined id.

*****************************************************************************/

int
neblo_register
( reqst_t *reqst )
{

  unsigned int          fport;
  u_int16_t             foreign_port;
  u_int32_t             foreign_regi_handle;
  char                 *foreign_keymod;
  char                 *foreign_keyexp;
  neblo_addr            foreign_id;
  neblo_addr            foreign_id_src;
  unsigned char        *n_A_k;
  unsigned char        *n_B_k;
  neblo_addr            n_A;
  neblo_addr            n_B;
  int                   rc;


  _do_(_malloc_(foreign_keymod,KEYBUFLEN+SLACK),
       !foreign_keymod,"neblo_register(): foreign_keymod malloc");
  _do_(_malloc_(foreign_keyexp,BUFLEN+SLACK),
       !foreign_keyexp,"neblo_register(): foreign_keyexp malloc");

  sscanf(reqst->buf,"%*s %u %u %50000s %100s",
         &fport,&foreign_regi_handle,foreign_keymod,foreign_keyexp);

  /*  SECURITY  */

  foreign_port = (u_int16_t)fport;
  if ((unsigned int)foreign_port != fport) {

#ifdef DEBUG
    S(fprintf(stderr,"[%lu]\tneblo_register():  bad foreign port.  Abort...\n",
              reqst->th_id));
#endif

    _free_(foreign_keymod); _free_(foreign_keyexp);
    return -1;
  }

  _do_(_malloc_(n_A_k,RIPEMD160_DIGEST_LENGTH+SLACK),
       !n_A_k,"neblo_register(): n_A_k malloc");
  _do_(_malloc_(n_B_k,RIPEMD160_DIGEST_LENGTH+SLACK),
       !n_B_k,"neblo_register(): n_B_k malloc");

  /* generate a 64 bit random number "B" */
  RAND_bytes((unsigned char *)&n_B,sizeof(n_B));

  /* get RIPEMD160(A) from foreign node */
  _read_(reqst->csock,n_A_k,RIPEMD160_DIGEST_LENGTH,rc,TIMEOUT,goto exit);

#ifdef DEBUG
  S(fprintf(stderr, "[%lu]\t>> (%u) %u %u %u %u %u\n",
          reqst->th_id,RIPEMD160_DIGEST_LENGTH,
          (unsigned int)n_A_k[0], (unsigned int)n_A_k[4],
          (unsigned int)n_A_k[8], (unsigned int)n_A_k[12],
          (unsigned int)n_A_k[16]));
#endif

  /* send RIPEMD160(B) to foreign node */
  RIPEMD160((unsigned char *)&n_B,sizeof(n_B),n_B_k);
  _write_(reqst->csock,n_B_k,RIPEMD160_DIGEST_LENGTH,rc,goto exit);

#ifdef DEBUG
  S(fprintf(stderr, "[%lu]\t<< (%u) %u %u %u %u %u\n",
          reqst->th_id,RIPEMD160_DIGEST_LENGTH,
          (unsigned int)n_B_k[0], (unsigned int)n_B_k[4],
          (unsigned int)n_B_k[8], (unsigned int)n_B_k[12],
          (unsigned int)n_B_k[16]));
#endif

  /* get A */
  _read_(reqst->csock,&n_A,sizeof(n_A),rc,TIMEOUT,goto exit);

#ifdef DEBUG
  S(fprintf(stderr,"[%lu]\t>> A was %016llx\n",reqst->th_id,n_A));
#endif
        
  /* check RIPEMD160(A) against the original "A". abort on fail */
  RIPEMD160((unsigned char *)&n_A,sizeof(n_A),n_B_k);
  if (strncmp((char *)n_A_k,(char *)n_B_k,RIPEMD160_DIGEST_LENGTH)) {

#ifdef DEBUG
    S(fprintf(stderr,"[%lu]\tcheck failed!\n",reqst->th_id));
#endif

    rc = -1;
    goto exit;
  }

  /* compute the unique id. of foreign node */
  foreign_id_src = n_A ^ n_B;  /*  this is the id. source  */
  RIPEMD160((unsigned char *)&foreign_id_src,sizeof(foreign_id_src),n_B_k);
  xor_fold((int8_t *)&foreign_id,sizeof(foreign_id),
           (int8_t *)n_B_k,RIPEMD160_DIGEST_LENGTH);

#ifdef DEBUG
  S(fprintf(stderr,"[%lu]\tforeign id is %016llx\n",reqst->th_id,foreign_id));
#endif

  /* send B, so foreign can validate our honesty */
  _write_(reqst->csock,&n_B,sizeof(n_B),rc,goto exit);

#ifdef DEBUG
  S(fprintf(stderr,"[%lu]\t<< B was %016llx\n",reqst->th_id,n_B));
#endif

  /*
   *  Phase 1 has come to end.  Begin phase 2, simulating an incoming
   *  "join" request.
   */

  sprintf(reqst->buf,"join %016llx %016llx %lu %u %u %s %s",
          foreign_id,foreign_id_src,
          (unsigned long)reqst->peer_addr.s_addr,foreign_port,
          foreign_regi_handle,
          foreign_keymod,foreign_keyexp);

  _close_(&reqst->csock);

  rc = neblo_join(reqst);

exit:
  _free_(foreign_keymod); _free_(foreign_keyexp);
  _free_(n_A_k); _free_(n_B_k);
  return rc;

}
